#!/usr/bin/perl -w


$::Cheat = 0;
$::Version = '0.8';

=comment

##########################################################################
#
# PANG ZERO
# Copyright (C) 2006 by UPi <upi at sourceforge.net>
#
##########################################################################

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2, as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


##########################################################################
# TODO:
##########################################################################
* Color the border during death sequence
* Music, sound effects (game pause, level advance)
* More levels, game balance for multiple players
* Message while paused
* Smaller playfield height
* Better joystick support
* "GAME OVER" message

/
=cut

use strict;
use SDL;
use SDL::App;
use SDL::Event;
use SDL::Surface;
use SDL::Timer;
use SDL::Palette;
use SDL::Sound;
use SDL::Mixer;
use SDL::Font;



# SDL objects

use vars qw ( 
  $App $Background $ScoreFont
  $HexaSurface $DarkHexaSurface
  $BallSurface $DarkBallSurface $GrayBallSurface $DarkGrayBallSurface
  $GreenBallSurface $DarkGreenBallSurface $GoldBallSurface $DarkGoldBallSurface $DeathBallSurface
  $BorderSurface $BonusSurface
  %Sounds $Mixer
);

# Pang Zero variables and objects

use vars qw (
  $ScreenHeight $ScreenWidth $PhysicalScreenWidth $PhysicalScreenHeight $ScreenMargin
  @GameObjects %GameEvents $PlayersAlive $GamePause 
  $GameLevel $GameSpeed $GameLevelDesc $GameLevelAdvance
  $SpawnDelay $SuperBallDelay $BlackBallDelay
  @Players $NumGuys @BallDesc @GameLevels
  %Keys %Events );
  
##########################################################################
# GLOBAL CONFIGURATION
##########################################################################

%Sounds = (
  'pop' => 'pop.voc',
  'shoot' => 'shoot.voc',
  'death' => 'meow.voc',
  'level' => 'level.voc',
  'bonuslife' => 'magic.voc',
  'pause' => 'pop3.voc',
);

$NumGuys = 2;
@Players = ( 
  { 'keys'  => [SDLK_LEFT, SDLK_RIGHT, SDLK_UP], 'hue' => 170, 'saturation' => 255 }, # blue
  { 'keys'  => [SDLK_a, SDLK_d, SDLK_s], 'hue' => 0, 'saturation' => 255 },           # red
  { 'keys'  => [SDLK_j, SDLK_l, SDLK_k], 'hue' => 85, 'saturation' => 255 },         # green
  { 'keys'  => [SDLK_KP6, SDLK_KP4, SDLK_KP5], 'hue' => 212, 'saturation' => 255 },   # pink
  { 'keys'  => [SDLK_KP6, SDLK_KP4, SDLK_KP5], 'hue' => 42, 'saturation' => 255 },   # yellow
  { 'keys'  => [SDLK_KP6, SDLK_KP4, SDLK_KP5], 'hue' => 128, 'saturation' => 255 },   # cyan
  { 'keys'  => [SDLK_KP6, SDLK_KP4, SDLK_KP5], 'hue' => 128, 'saturation' => 0 },   # gray
  { 'keys'  => [SDLK_KP6, SDLK_KP4, SDLK_KP5], 'hue' => 113, 'saturation' => 128 },   # snot
  { 'keys'  => [SDLK_KP6, SDLK_KP4, SDLK_KP5], 'hue' => 212, 'saturation' => 64 },   # purple
);

@BallDesc = (
# Normal balls (n0 .. n4)
  { 'name' => 'n0', 'score' =>  2000, 'spawndelay' =>   1, 'width' => 128, 'height' =>106, 'speedY' => 6.5, 'popIndex' => 0,
    'rect' => new SDL::Rect(-x => 0, -y => 12, -width => 128, -height => 106) },
  { 'name' => 'n1', 'score' =>  1000, 'spawndelay' => 0.5, 'width' =>  96, 'height' => 80, 'speedY' => 5.7, 'popIndex' => 1,
    'rect' => new SDL::Rect(-x => 128, -y => 8, -width => 96, -height => 80) },
  { 'name' => 'n2', 'score' =>   800, 'spawndelay' => 0.25, 'width' =>  64, 'height' => 53, 'speedY' => 5, 'popIndex' => 2,
    'rect' => new SDL::Rect(-x => 0, -y => 134, -width => 64, -height => 53) },
  { 'name' => 'n3', 'score' =>   600, 'spawndelay' => 0.12, 'width' =>  32, 'height' => 28, 'speedY' => 4, 'popIndex' => 3,
    'rect' => new SDL::Rect(-x => 224, -y => 2, -width => 32, -height => 28) },
  { 'name' => 'n4', 'score' =>   500, 'spawndelay' => 0.05, 'width' =>  16, 'height' => 15, 'speedY' => 3, 'popIndex' => 4,
    'rect' => new SDL::Rect(-x => 232, -y => 42, -width => 16, -height => 15) },
# "Bouncy" balls (b0..b2)
  { 'name' => 'b0', 'score' =>  1500, 'spawndelay' => 0.5, 'width' =>  64, 'height' => 52, 'speedY' => 5.7, 'popIndex' => 2,
    'rect' => new SDL::Rect(-x => 160, -y => 166, -width => 64, -height => 52) },
  { 'name' => 'b1', 'score' =>   750, 'spawndelay' => 0.2, 'width' =>  32, 'height' => 28, 'speedY' => 5, 'popIndex' => 3,
    'rect' => new SDL::Rect(-x => 224, -y => 66, -width => 32, -height => 28) },
  { 'name' => 'b2', 'score' =>   500, 'spawndelay' => 0.1, 'width' =>  16, 'height' => 14, 'speedY' => 4.2, 'popIndex' => 4,
    'rect' => new SDL::Rect(-x => 232, -y => 106, -width => 16, -height => 14) },
# Hexas (h0..h2)
  { 'name' => 'h0', 'score' =>  1500, 'spawndelay' => 0.5, 'width' =>  64, 'height' => 52, 'popIndex' => 5, 'hexa' => 1,
    'rect' => new SDL::Rect(-x => 0, -y => 6, -width => 64, -height => 52),
    'magicrect' => new SDL::Rect(-x => 0, -y => 102, -width => 64, -height => 52), },
  { 'name' => 'h1', 'score' =>  1000, 'spawndelay' => 0.2, 'width' =>  32, 'height' => 28, 'popIndex' => 6, 'hexa' => 1,
    'rect' => new SDL::Rect(-x => 0, -y => 66, -width => 32, -height => 28),
    'magicrect' => new SDL::Rect(-x => 0, -y => 162, -width => 32, -height => 28), },
  { 'name' => 'h2', 'score' =>   500, 'spawndelay' => 0.1, 'width' =>  16, 'height' => 14, 'popIndex' => 7, 'hexa' => 1,
    'rect' => new SDL::Rect(-x => 192, -y => 65, -width => 16, -height => 14),
    'magicrect' => new SDL::Rect(-x => 9, -y => 200, -width => 16, -height => 14), },
# Death
  { 'name' => 'death', 'score' => 0, 'spawndelay' => 0.1, 'width' =>  64, 'height' => 53, 'speedY' => 5, 'popIndex' => 2,
    'rect' => new SDL::Rect(-x => 0, -y => 134, -width => 64, -height => 53) },
);
$BallDesc[0]->{nextgen} = $BallDesc[1];
$BallDesc[1]->{nextgen} = $BallDesc[2];
$BallDesc[2]->{nextgen} = $BallDesc[3];
$BallDesc[3]->{nextgen} = $BallDesc[4];

$BallDesc[5]->{nextgen} = $BallDesc[6];
$BallDesc[6]->{nextgen} = $BallDesc[7];

$BallDesc[8]->{nextgen} = $BallDesc[9];
$BallDesc[9]->{nextgen} = $BallDesc[10];

$BallDesc[11]->{nextgen} = $BallDesc[11];

my %BallMixes;
%BallMixes = (
  'easy'   => [ qw(n0 2 n1 20 n2 10 n3 3 n4 2   b0 5 b1 2 b2 1   h0 5 h1 3 h2 1) ],
  'medium' => [ qw(n0 10 n1 20 n2 10 n3 3 n4 2  b0 10 b1 2 b2 1  h0 15 h1 5 h2 1  death 2) ],
  'bouncy' => [ qw(n0 20 n1 10 n2 5 n3 1 n4 1   b0 30 b1 9 b2 1  h0 15 h1 5       death 5) ],
  'hard'   => [ qw(n0 20 n1 10 n2 5 n3 1        b0 20 b1 2       h0 20 h1 5       death 10) ],
  'hexas'  => [ qw(n0 10 n1 10 n2 5 n3 1        b0 15 b1 2       h0 40 h1 15      death 5) ],
);

sub AddLevels {
  my ($num, $balls, $gamespeedStart, $gamespeedEnd, $spawndelayStart, $spawndelayEnd) = @_;
  my ($i, $level);
  
  for ($i = 0; $i < $num; ++$i) {
    $level = {
      'balls' => $balls,
      'gamespeed' => $gamespeedStart + ($gamespeedEnd - $gamespeedStart) * ($i) / ($num), 
      'spawndelay' => $spawndelayStart + ($spawndelayEnd - $spawndelayStart) * ($i) / ($num), 
    };
    push @GameLevels, ( $level );
  }
}

&AddLevels(  9, $BallMixes{easy},   0.75, 1.25, 20, 20 ); # 0-9
&AddLevels( 10, $BallMixes{medium}, 0.7 , 1.3 , 20, 15 ); # 1x
&AddLevels( 10, $BallMixes{hard},   0.7 , 1.5 , 15, 15 ); # 2x
&AddLevels( 10, $BallMixes{hexas},  1.0 , 1.5 , 15, 12 ); # 3x
&AddLevels( 10, $BallMixes{hard},   0.7 , 1.7 , 15, 17 ); # 4x
&AddLevels( 10, $BallMixes{bouncy}, 1.0 , 2.0 , 12, 12 ); # 5x
&AddLevels( 10, $BallMixes{medium}, 1.5 , 2.2 , 13,  8 ); # 6x
&AddLevels( 10, $BallMixes{hard},   1.0 , 2.2 , 13, 10 ); # 7x
&AddLevels( 10, $BallMixes{hexas},  1.3 , 2.4 , 12,  9 ); # 8x
&AddLevels( 10, $BallMixes{hard},   2.0 , 3.0 , 13, 10 ); # 9x

$PhysicalScreenWidth = 800;
$PhysicalScreenHeight = 600;
$ScreenMargin = 16;
$ScreenWidth = $PhysicalScreenWidth - $ScreenMargin * 2;
$ScreenHeight = int( ($PhysicalScreenHeight - $ScreenMargin * 2 - 150) / 32 ) * 32;


##########################################################################
# GAME OBJECT CLASSES
##########################################################################

package GameObject;
package Ball;
package Hexa;
package SuperBall;
package DeathBall;
package DarkBall;
package Pop;
package GamePause;
package Guy;
package Harpoon;
package DeadGuy;


##########################################################################
package GameObject;
##########################################################################

sub new {
  my ($class) = @_;
  my $self = {
    'rect' => new SDL::Rect( -x => 0, -y => 0, -width => 0, -height => 0 ),
    'speedX' => 0,
    'speedY' => 0,
    'x' => 0,
    'y' => 0,
    'w' => 10,
    'h' => 10,
  };
  bless $self, $class;
}

sub Delete {
  my $self = shift;
  my ($i);
  
  for ($i = 0; $i < scalar @::GameObjects; ++$i) {
    if ($::GameObjects[$i] eq $self) {
      splice @::GameObjects, $i, 1;
      last;
    }
  }
  $self->{deleted} = 1;
  $self->Clear();
}

sub Advance {
}

sub Clear {
  my ($self) = @_;
  $::Background->blit($self->{rect}, $::App, $self->{rect});
  #$::App->fill( $self->{rect}, new SDL::Color() );
}

sub TransferRect {
  my ($self) = @_;
  
  $self->{rect}->x($self->{x} + $::ScreenMargin);
  $self->{rect}->y($self->{y} + $::ScreenMargin);
  $self->{rect}->width($self->{w});
  $self->{rect}->height($self->{h});
}

sub Draw {
  my ($self) = @_;
  
  $self->TransferRect();
  $::App->fill( $self->{rect}, new SDL::Color(-r => 0x80) );
}



##########################################################################
package Ball;
##########################################################################

@Ball::ISA = qw(GameObject);
$Ball::Gravity = 0.05;
$Ball::MagicBallRect = new SDL::Rect(-x => 232, -y => 138, -width => 16, -height => 15);

for (my $i=0; $i <= $#::BallDesc; ++$i) {
  my $desc = $::BallDesc[$i];
  next if $desc->{hexa};
  $desc->{bounceY} = $desc->{speedY} * $desc->{speedY} / $Ball::Gravity / 2;
}

sub Create {
  my ($description, $x, $y, $dir, $superball) = @_;
  return new SuperBall(@_) if $superball;
  return new DeathBall(@_) if $description->{name} eq 'death';
  return new Hexa(@_) if ($description->{hexa});
  return new Ball(@_);
}


sub new {
  my ($class, $description, $x, $y, $dir) = @_;
  my ($self);
  
  $self = new GameObject;
  %{$self} = ( %{$self},
    'x' => $x,
    'y' => $y,
    'w' => $description->{width},
    'h' => $description->{height},
    'surface' => $::BallSurface,
    'anim' => 0,
    'hexa' => $description->{hexa} ? 1 : 0,
    'desc' => $description,
    'hasmagic' => 0,		# true if one of the ball's descendants is magic
    'ismagic' => 0,			# true if the ball IS magic
  );
  $self->{speedX} = $dir > 0 ? 1.3 : -1.3;
  
  bless $self, $class;
}

sub Advance {
  my ($self) = @_;
  
  ++$self->{anim};
  unless( $::GamePause > 0 ) {
    $self->{speedY} += $Ball::Gravity * $::GameSpeed unless ($self->{hexa});
    $self->{x} += $self->{speedX} * $::GameSpeed;
    $self->{y} += $self->{speedY} * $::GameSpeed;
    if ($self->{y} > $::ScreenHeight - $self->{h}) {
        $self->{y} = $::ScreenHeight - $self->{h};
      if ($self->{hexa}) {
        $self->{speedY} = -abs($self->{speedY});
      } else {
        $self->{speedY} = -$self->{desc}->{speedY};
      }
      $self->Bounce;
    }
    if ($self->{y} < 0) {
      $self->{y} = 0;
      $self->{speedY} = abs($self->{speedY});
    }
    if ($self->{x} < 0) {
      $self->{x} = 0;
      $self->{speedX} = abs( $self->{speedX} );
    }
    if ($self->{x} > $::ScreenWidth - $self->{w}) {
      $self->{x} = $::ScreenWidth - $self->{w};
      $self->{speedX} = -abs( $self->{speedX} );
    }
  }
  
  $self->CheckCollisions();
}

sub Bounce {
}

sub CheckCollisions {
  my $self = shift;  
  my ($gameObject);
  
  foreach $gameObject (@::GameObjects) {
    if ( ref $gameObject eq 'Harpoon' ) {
      if ($self->Collisions($gameObject)) {
        $self->Pop($gameObject->{guy}, '');
        $gameObject->Delete();
        last;
      }
    }
    elsif ( ref $gameObject eq 'Guy' ) {
      if ($::GamePause <= 0 and $self->Collisions($gameObject)) {
        $gameObject->Kill();
      }
    }
  }
}

sub Draw {
  my ($self) = @_;
  
  $self->TransferRect();
  if ($self->{ismagic} and int($self->{anim}/4) % 2) {
    $self->{surface}->blit( $Ball::MagicBallRect, $::App, $self->{rect} )
  } else {
    $self->{surface}->blit( $self->{desc}->{rect}, $::App, $self->{rect} );
  }
}

sub Collisions {
  my ($self, $other) = @_;
  
  # Bounding box detection
  
  return 0 if $self->{x} >= $other->{x} + $other->{w};
  return 0 if $other->{x} >= $self->{x} + $self->{w};
  return 0 if $self->{y} >= $other->{y} + $other->{h};
  return 0 if $other->{y} >= $self->{y} + $self->{h};
  
  # Circle vs rectangle collision
  
  my ($centerX, $centerY, $boxAxisX, $boxAxisY, $boxCenterX, $boxCenterY, $distSquare, $distance);
  $boxAxisX = ($other->{collisionw} or $other->{w}) / 2;
  $boxAxisY = ($other->{collisionh} or $other->{h}) / 2;
  $boxCenterX = $other->{x} + $other->{w} / 2;
  $boxCenterY = $other->{y} + $other->{h} / 2;
  $centerX = $self->{x} + $self->{w} / 2;
  $centerY = $self->{y} + $self->{h} / 2;
  
  # Translate coordinates to the box center
  $centerX -= $boxCenterX;
  $centerY -= $boxCenterY;  
  $centerX = abs($centerX);
  $centerY = abs($centerY);
  
  if ($centerX < $boxAxisX) {
    return 1 if $centerY < $boxAxisY + $self->{h} / 2;
    return 0;
  }
  if ($centerY < $boxAxisY) {
    return 2 if $centerX < $boxAxisX + $self->{w} / 2;
    return 0;
  }
  $distSquare = ($centerX-$boxAxisX) * ($centerX-$boxAxisX);
  $distSquare+= ($centerY-$boxAxisY) * ($centerY-$boxAxisY);
  return 3 if $distSquare < $self->{w} * $self->{h} / 4;
  
  return 0;
}

sub Pop {
  my ($self, $guy, $superkill) = @_;
  
  &::PlaySound('pop') unless $superkill eq 'superkill';
  $::GameEvents{'magic'} = 1 if ($self->{ismagic});
  $guy->GiveScore($self->{desc}->{score}) if $guy;
  
  if ($self->{desc}->{nextgen}) {
    my @children = $self->SpawnChildren($superkill);
    if (scalar @children) {
      $self->AdjustChildren(@children);
      push @::GameObjects, (@children);
    }
  }
  push @::GameObjects, (new Pop($self->{x}, $self->{y}, $self->{desc}->{popIndex}, $self->{surface}));
  &::OnBallPopped();
  $self->Delete();
}

sub SpawnChildren {
  my $self = shift;
  my ($nextgen, $child1, $child2, $x, $y);
  
  $nextgen = $self->{desc}->{nextgen};
  $x = $self->{x} + $self->{w} / 2;
  $y = $self->{y} + ( $self->{h} - $nextgen->{height} ) / 2;

  $child1 = &Create($nextgen, $self->{x}, $y, 0);
  $child2 = &Create($nextgen, $self->{x} + $self->{w} - $nextgen->{width}, $y, 1);
  if ($self->{hasmagic}) {
    (rand(50) > 25 ? $child1 : $child2)->GiveMagic();
  }
  return ($child1, $child2);
}

sub AdjustChildren {
  my $self = shift;
  my ($nextgen, $child1, $child2, $speedY, $altitude);
  
  $nextgen = $self->{desc}->{nextgen};
  $altitude = $::ScreenHeight - $self->{y} - $self->{h};
  if ($altitude > $nextgen->{bounceY}) {
    $speedY = 1.8;
  } else {
    $speedY = 1.8;
    while ($speedY * $speedY / $Ball::Gravity / 2 + $altitude < $nextgen->{bounceY}) {
      ++$speedY;
    }
  }

  $_[0]->{speedY} = -$speedY;
  $_[1]->{speedY} = -$speedY;
}

sub GiveMagic {
  my $self = shift;
  
  $self->{hasmagic} = 1;
  $self->{ismagic} = 1 unless $self->{desc}->{nextgen};
}



##########################################################################
package Hexa;
##########################################################################

@Hexa::ISA = qw(Ball);

sub new {
  my $class = shift;
  my ($self);
  
  $self = new Ball(@_);
  $self->{surface} = $::HexaSurface;
  $self->{speedX} = (rand(1.25) + 1.25) * ($self->{speedX} > 0 ? 1 : -1);
  $self->{speedY} = -4 + abs($self->{speedX});
  
  bless $self, $class;
}

sub Draw {
  my $self = shift;
  my ($rect, $srcx);
  
  $self->TransferRect();
  if ($self->{ismagic} and int($self->{anim} / 3) % 3 == 0) {
    $self->{surface}->blit($self->{desc}->{magicrect}, $::App, $self->{rect});
  } else {
    $rect = $self->{desc}->{rect};
    $srcx = (int($self->{anim} / 3) % 3) * $self->{w};
    $rect->x( $rect->x + $srcx );
    $self->{surface}->blit( $rect, $::App, $self->{rect} );
    $rect->x( $rect->x - $srcx );
  }
}

sub AdjustChildren {
}


##########################################################################
package SuperBall;
##########################################################################

@SuperBall::ISA = qw(Ball);

sub new {
  my $class = shift;
  my ($self);
  
  $self = new Ball(@_);
  $self->{effect} = 0;   # 0 : superpause;  1 : superkill
  $self->{surface} = $self->{effect} ? $::GoldBallSurface : $::GreenBallSurface;
  
  bless $self, $class;
}

sub Bounce {
  my $self = shift;
  $self->{effect} = 1 - $self->{effect};
  $self->{surface} = $self->{effect} ? $::GoldBallSurface : $::GreenBallSurface;
}

sub SpawnChildren {
  return ();
}

sub Pop {
  my ($self) = shift;
  
  $self->SUPER::Pop(@_);
  if ($self->{effect} == 0) {
    $::GameEvents{superpause} = 1;
  } else {
    $::GameEvents{superkill} = $_[0];  # The superkiller guy
  }
}


##########################################################################
package DeathBall;
##########################################################################

@DeathBall::ISA = qw(Ball);

sub new {
  my $class = shift;
  my ($self);
  
  $self = new Ball(@_);
  $self->{surface} = $::DeathBallSurface;
  $self->{expires} = 2000;   # 20sec
  $self->{speedX} *= 0.9;
  bless $self, $class;
}

sub Advance {
  my $self = shift;
  
  $self->SUPER::Advance();
  if (--$self->{expires} < 0) {
    $self->Pop(undef, 'expire');
  }
  
}

sub SpawnChildren {
  my $self = shift;
  my ($superkill) = @_;
  
  return if $superkill;
  $self->SUPER::SpawnChildren(@_);
}


##########################################################################
package DarkBall;
##########################################################################

@DarkBall::ISA = qw(GameObject);

sub new {
  my ($class, $description, $x, $dir, $superball) = @_;
  
  my $self = new GameObject;
  %{$self} = ( %{$self},
    'x' => $x,
    'y' => -$description->{height},
    'w' => $description->{width},
    'h' => $description->{height},
    'desc' => $description,
    'delay' => 0,
    'dir' => $dir,
    'superball' => $superball,
  );
  bless $self, $class;
}

sub Advance {
  my $self = shift;
  
  return if $::GamePause > 0;
  --$self->{delay};
  if ($self->{delay} < 0) {
    $self->{y} += 16;
    $self->{delay} = 50 / $::GameSpeed;
    if ($self->{y} >= 0) {
      my $ball = Ball::Create( $self->{desc}, $self->{x}, $self->{y}, $self->{dir}, $self->{superball} );
      $ball->GiveMagic;
      push @::GameObjects, ( $ball );
      $self->Delete();
    }
  }
}

sub Draw {
  my $self = shift;
  my ($surface);
  
  $self->TransferRect();
  $surface = $self->{desc}->{hexa} ? $::DarkHexaSurface : $::DarkBallSurface;
  $surface = $::DarkGreenBallSurface if $self->{superball};
  $surface = $::DeathBallSurface if $self->{desc}->{name} eq 'death';
  $surface->blit( $self->{desc}->{rect}, $::App, $self->{rect} );
}



##########################################################################
package Pop;
##########################################################################

@Pop::ISA = qw(GameObject);

@Pop::Description = (
  { 'hexa' => 0, 'xoffset' => 0, 'yoffset' => 12, 'srcx' => 256, 'srcy' =>   0, 'size' => 128, },
  { 'hexa' => 0, 'xoffset' => 0, 'yoffset' =>  8, 'srcx' => 256, 'srcy' => 128, 'size' =>  96, },
  { 'hexa' => 0, 'xoffset' => 0, 'yoffset' =>  6, 'srcx' => 256, 'srcy' => 224, 'size' =>  64, },
  { 'hexa' => 0, 'xoffset' => 0, 'yoffset' =>  2, 'srcx' => 512, 'srcy' => 224, 'size' =>  32, },
  { 'hexa' => 0, 'xoffset' =>-8, 'yoffset' => -8, 'srcx' => 512, 'srcy' => 256, 'size' =>  32, },
  
  { 'hexa' => 1, 'xoffset' => 0, 'yoffset' => -6, 'srcx' =>   0, 'srcy' =>  96, 'size' =>  64, },
  { 'hexa' => 1, 'xoffset' => 0, 'yoffset' => -4, 'srcx' =>   0, 'srcy' => 160, 'size' =>  32, },
  { 'hexa' => 1, 'xoffset' =>-8, 'yoffset' => -6, 'srcx' =>   0, 'srcy' => 192, 'size' =>  32, },
);

sub new {
  my ($class, $x, $y, $index, $surface) = @_;
  my ($self, $desc);
  
  $desc = $Pop::Description[$index],
  $self = new GameObject;
  %{$self} = ( %{$self},
    'x' => $x + $desc->{xoffset},
    'y' => $y + $desc->{yoffset},
    'w' => $desc->{size},
    'h' => $desc->{size},
    'desc' => $desc,
    'anim' => 0,
    'surface' => $surface,
  );
  bless $self, $class;
}

sub Advance {
  my $self = shift;
  
  ++$self->{anim};
  if ($self->{anim} >= 20) {
    $self->Delete();
  }
}

sub Draw {
  my $self = shift;
  my ($phase, $srcrect);
  
  $self->TransferRect();
  $phase = int($self->{anim} / 5);
  $phase = 3 if $phase > 3;
  
  $srcrect = new SDL::Rect(
    -x => $self->{desc}->{srcx} + $phase * $self->{w},
    -y => $self->{desc}->{srcy},
    -width => $self->{w},
    -height => $self->{h} );
  $self->{surface}->blit( $srcrect, $::App, $self->{rect} );
}


##########################################################################
package GamePause;
##########################################################################

@GamePause::ISA = qw(GameObject);

use vars qw($Shown);

sub Show {
  return if $Shown;
  $Shown = 1;
  unshift @::GameObjects, (new GamePause);
}

sub new {
  my ($class) = @_;
  
  my $self = new GameObject;
  my ($width);
  $width = ::SDL_TEXTWIDTH("Time left: 9.999");
  
  %{$self} = ( %{$self},
    'x' => ($::PhysicalScreenWidth - $width) / 2,
    'y' => 100,
    'w' => $width,
    'h' => 32,
  );
  $self->TransferRect();
  bless $self, $class;
}

sub Delete {
  my $self = shift;
  
  $self->SUPER::Delete();
  $Shown = 0;
}

sub Advance {
  my $self = shift;
  
  if ($::GamePause <= 0) {
    $self->Delete;
  }
}

sub Draw {
  my $self = shift;
  
  $::App->print( $self->{rect}->x, $self->{rect}->y, "Time left: " . ($::GamePause / 100) );
}





##########################################################################
package Guy;
##########################################################################

@Guy::ISA = qw(GameObject);

sub new {
  my ($class, $number) = @_;
  my ($self, $desc);
  
  $self = new GameObject;
  $desc = $::Players[$number];

  %{$self} = ( %{$self},
    'number' => $number,
    'x' => $desc->{startX},
    'y' => $::ScreenHeight - 64,
    'w' => 64,
    'h' => 64,
    'collisionw' => '30',
    'collisionh' => '55',
    'delay' => 0,
    'speedY' => 0,
    'speedX' => 0,
    'dir' => $number % 2,
    'state' => 'idle',
    'killed' => 0,
    'harpoons' => 0,
    'invincible' => 0,
    'desc' => $desc,
    'surface' => $desc->{guySurface},
  );
  bless $self, $class;
}

sub Fire {
  my ($self) = @_;
  
  if ($self->{harpoons} < 2) {
    ++$self->{harpoons};
    unshift @::GameObjects, (new Harpoon($self));
    $self->{state} = 'shoot';
    $self->{delay} = 7;
    ::PlaySound('shoot');
    return 1;
  }
  return 0;
}

sub Advance {
  my ($self) = @_;
  
  --$self->{invincible};
  
  if ($self->{delay} > 0) {
    --$self->{delay};
    return;
  }
  
  $self->{speedX} = 0;
  $self->{state} = 'idle';
  my $keys = $self->{desc}->{keys};
  
  if ( $::Events{$keys->[2]} ) {
    return if $self->Fire();
  }
  if ( $::Keys{$keys->[0]} ) {
    $self->{speedX} = -3;
    $self->{dir} = 0;
    $self->{state} = 'walk';
  } elsif ( $::Keys{$keys->[1]} ) {
    $self->{speedX} = 3;
    $self->{dir} = 1;
    $self->{state} = 'walk';
  }
  $self->{x} += $self->{speedX};
  
  if ($self->{x} < -16) {
    $self->{x} = -16;
  }
  if ($self->{x} > $::ScreenWidth - $self->{w} + 16) {
    $self->{x} = $::ScreenWidth - $self->{w} + 16;
  }
}

sub Draw {
  my ($self) = @_;
  my ($srcrect, $srcx, $srcy, $srcw, $srch);
  
  return if ($self->{killed});
  return if ($self->{invincible} > 0 and int($self->{invincible} / 4) % 2);
  
  $srcw = $srch = 64;
  if ($self->{state} eq 'idle') {
    $srcx = $self->{dir} * 128;
    $srcy = 64;
  } elsif ($self->{state} eq 'walk') {
    $srcx = $self->{dir} * 256 + (int($self->{x} / 16) % 4) * 64;
    $srcy = 0;
  } elsif ($self->{state} eq 'shoot') {
    $srcx = $self->{dir} * 128 + 64;
    $srcx -= 64 if ($self->{delay} <= 1);
    $srcy = 64;
  }
  $srcrect = new SDL::Rect( -width => $srcw, -height => $srch, -x => $srcx, -y => $srcy );
  $self->TransferRect();
  $self->{surface}->blit($srcrect, $::App, $self->{rect});
}

sub Kill {
  my ($self) = @_;
  
  return if $::Cheat;
  return if $self->{invincible} > 0;
  $self->{justkilled} = 1;
  $::GameEvents{'kill'} = 1;
}

sub DeleteHarpoons {
  my ($self) = @_;
  my (@gameObjects, $harpoon);
  
  @gameObjects = @::GameObjects;
  foreach my $harpoon (@gameObjects) {
    $harpoon->Delete if ($harpoon->{guy} and $harpoon->{guy} eq $self);
  }
}

sub GiveScore {
  my ($self, $score) = @_;
  
  $self->{desc}->{score} += $score;
  if ($self->{desc}->{score} > $self->{desc}->{scoreforbonuslife}) {
    ++$self->{desc}->{lives};
    $self->{desc}->{scoreforbonuslife} += 200000;
    &::PlaySound('bonuslife');
  }
}



##########################################################################
package Harpoon;
##########################################################################

@Harpoon::ISA = qw(GameObject);

sub new {
  my ($class, $guy) = @_;
  my ($self);
  
  $self = new GameObject;
  %{$self} = ( %{$self},
    'x' => $guy->{x} + 22,
    'y' => $::ScreenHeight - 32,
    'w' => 18,
    'h' => 32,
    'anim' => 0,
    'speedY' => -3,
    'speedX' => 0,
    'guy' => $guy,
    'surface' => $guy->{desc}->{harpoonSurface},
  );
  bless $self, $class;
}

sub Delete {
  my $self = shift;
  
  --$self->{guy}->{harpoons};
  $self->SUPER::Delete();
}

sub Advance {
  my $self = shift;
  
  $self->{y} += $self->{speedY};
  if ($self->{y} < 0) {
    $self->Delete();
    return;
  }
  $self->{h} = $::ScreenHeight - $self->{y};
  ++$self->{anim};
}

sub Draw {
  my $self = shift;
  my ($x, $y, $h, $maxh, $dstrect, $srcrect);
  
  $self->TransferRect();
  $y = $self->{y};
  $dstrect = new SDL::Rect( -w => $self->{w}, -x => $self->{x} + $::ScreenMargin );
  $srcrect = new SDL::Rect( -w => $self->{w}, 
    -x => (40, 70, 102)[ int($self->{anim} / 4) % 3 ], -y => 0 );
  $maxh = 64;
  
  # The harpoon needs to be drawn from tile pieces.
  # $y iterates from $self->{y} to $::ScreenHeight
  # We draw at most $maxh height tiles at a time.
  
  while ($y < $::ScreenHeight) {
    $h = $::ScreenHeight - $y;
    $h = $maxh if $h > $maxh;
    $dstrect->y( $y + $::ScreenMargin );
    $dstrect->height( $h );
    $srcrect->height( $h );
    $self->{surface}->blit( $srcrect, $::App, $dstrect );
    
    # Prepare for next piece
    $y += $h;
    $srcrect->y( 32 ); # First piece starts at 0, rest start at 32
    $maxh = 32;        # First piece maxheight is 64, rest is 32
  }
}



##########################################################################
package DeadGuy;
##########################################################################

@DeadGuy::ISA = qw(GameObject);

sub new {
  my ($class, $guy, $dir) = @_;
  my ($self, $number, $desc);
  
  $self = new GameObject;
  $number = $guy->{number};
  $desc = $guy->{desc};

  %{$self} = ( %{$self},
    'number' => $number,
    'x' => $guy->{x},
    'y' => $guy->{y},
    'w' => 64,
    'h' => 64,
    'speedY' => -7,
    'surface' => $desc->{guySurface},
  );
  $self->{'speedX'} = 3 * (($self->{x} > $::ScreenWidth / 2) ? 1 : -1);
  bless $self, $class;
}

sub Advance {
  my $self = shift;
  
  $self->{speedY} += 0.1;
  $self->{x} += $self->{speedX};
  $self->{y} += $self->{speedY};
    
  if ($self->{x} < -16) {
    $self->{x} = -16;
    $self->{speedX} = abs( $self->{speedX} );
  }
  if ($self->{x} > $::ScreenWidth - $self->{w} +16) {
    $self->{x} = $::ScreenWidth - $self->{w} + 16;
    $self->{speedX} = -abs( $self->{speedX} );
  }
  
  if ($self->{y} > $::PhysicalScreenWidth) {
    $self->Delete;
  }
}

sub Draw {
  my $self = shift;
  my ($srcrect);
  
  $srcrect = new SDL::Rect(
    -x => ($self->{number} % 2) * 192 + ($self->{speedX} > 0 ? 0 : 64),
    -y => ($self->{number} % 2) * 64 + 128,
    -width => 64, -height => 64 );
  $self->TransferRect();
  $self->{surface}->blit( $srcrect, $::App, $self->{rect} );
}


##########################################################################
package MenuItem;
##########################################################################

@MenuItem::ISA = qw(GameObject);

sub new {
  my ($class, $x, $y, $text) = @_;
  my ($self);
  
  $self = new GameObject;
  
  %{$self} = ( %{$self},
    'x' => $x,
    'y' => $y,
    'h' => 40,
    'selected' => 0,
    'filled' => 0,
    'fillcolor' => new SDL::Color(-b=>128),
  );
  bless $self, $class;
  $self->SetText($text);
  return $self;
}

sub Advance {
}

sub Clear {
  my $self = shift;

  if ($self->{selected} or $self->{filled}) {
    $::App->fill($self->{rect}, $self->{fillcolor});
  } else {
    $self->SUPER::Clear();
  }
}

sub Draw {
  my $self = shift;
  
  $::App->print($self->{x} + 5 +$::ScreenMargin, $self->{y} + $::ScreenMargin, $self->{text});
}

sub SetText {
  my ($self, $text) = @_;
  
  $self->SUPER::Clear();
  $self->{text} = $text;
  $self->{w} = ::SDL_TEXTWIDTH($text) + 10;
  $self->TransferRect();
}

sub Select {
  my ($self) = @_;
  
  foreach my $item (@::GameObjects) {
    $item->{selected} = 0 if ref $item eq 'MenuItem';
  }
  $self->{selected} = 1;
}



##########################################################################
package GameTimer;
##########################################################################

use vars qw($FirstTick $LastTick $TotalAdvances);

sub ResetTimer {
  $FirstTick = $::App->ticks;
  $LastTick = $FirstTick;
  $TotalAdvances = 0;
}

sub GetAdvances {
  my ($ticks, $advance);
  
  $ticks = $::App->ticks;
  $advance = int(($ticks - $FirstTick) / 10) - $TotalAdvances;
  $TotalAdvances += $advance;
  return $advance;
}



##########################################################################
# JOYSTICKS
##########################################################################

package Joystick;
use vars qw(@Joysticks);

sub InitJoystick {
  my $numJoysticks = &SDL::NumJoysticks();
  for (my $i = 0; $i < $numJoysticks; ++$i) {
    print STDERR "Found joystick " , $i+1 , ": " , &SDL::JoystickName($i), "\n";
    push @Joysticks, &SDL::JoystickOpen($i);
  }
}

sub ReadJoystick {
  my ($i);
  
  $i = 0;
  foreach my $joystick (@Joysticks) {
    ++$i;
    my $axis = &SDL::JoystickGetAxis($joystick, 0);
    if ($axis <= -128) {
      $::Events{"L$i"} = 1 unless $::Keys{"L$i"};
      $::Keys{"L$i"} = 1;
      $::Keys{"R$i"} = 0;
    } elsif ($axis >= 128) {
      $::Events{"R$i"} = 1 unless $::Keys{"R$i"};
      $::Keys{"R$i"} = 1;
      $::Keys{"L$i"} = 0;
    } else {
      $::Keys{"L$i"} = 0;
      $::Keys{"R$i"} = 0;
    }
    if (&SDL::JoystickGetButton($joystick, 0)) {
      $::Events{"B$i"} = 1 unless $::Keys{"B$i"};
      $::Keys{"B$i"} = 1;
    } else {
      $::Keys{"B$i"} = 0;
    }
  }
}


##########################################################################
# PALETTE MANIPULATION
##########################################################################

package main;

sub RgbToHsi {
  my ($r, $g, $b) = @_;
  my ($min, $max, $delta, $h, $s, $i);
  
  if ($r > $g) { 
    $max = $r > $b ? $r : $b;
    $min = $g < $b ? $g : $b;
  } else {
    $max = $g > $b ? $g : $b;
    $min = $r < $b ? $r : $b;
  }
  $i = ($min + $max) / 2;
  if ($min == $max) {
    return (0, 0, $i);
  }
  
  $delta = ($max - $min);
  if ($i < 128) {
    $s = 255 * $delta / ($min + $max);
  } else {
    $s = 255 * $delta / (511 - $min - $max);
  }
  
  if ($r == $max) {
    $h = ($g - $b) / $delta;
  } elsif ($g == $max) {
    $h = 2 + ($b - $r) / $delta;
  } else {
    $h = 4 + ($r - $g) / $delta;
  }
  $h = $h * 42.5;
  $h += 255 if $h < 0;
  $h -= 255 if $h > 255;
  
  return ($h, $s, $i);
}

sub HsiToRgb {
  my ($h, $s, $i) = @_;
  my ($m1, $m2);
  
  if ($s < 1) {
    $i = int($i + 0.5);
    return ($i, $i, $i);
  }
  
  if ($i < 128) {
    $m2 = ($i * (255 + $s)) / 65025.0;
  } else {
    $m2 = ($i + $s - ($i * $s) / 255.0) / 255.0;
  }
  $m1 = ($i / 127.5) - $m2;

  return (
    &GetHsiValue( $m1, $m2, $h + 85),
    &GetHsiValue( $m1, $m2, $h),
    &GetHsiValue( $m1, $m2, $h - 85)
  );
}

sub GetHsiValue {
  my ($n1, $n2, $hue) = @_;
  my ($value);
  
  $hue -= 255 if ($hue > 255);
  $hue += 255 if ($hue < 0);
  if ($hue < 42.5) {
    $value = $n1 + ($n2 - $n1) * ($hue / 42.5);
  } elsif ($hue < 127.5) {
    $value = $n2;
  } elsif ($hue < 170) {
    $value = $n1 + ($n2 - $n1) * ((170 - $hue) / 42.5);
  } else {
    $value = $n1;
  }
  
#  print "\nGetHsiValue($n1, $n2,$hue) = ", $value*255, "\n";

  return int($value * 255 + 0.5);
}

sub LoadSurfaces {
  my ($i);
  $BallSurface = new SDL::Surface( -name => "ball.png" );
  $DarkBallSurface = new SDL::Surface( -name => "ball.png" );
  $GreenBallSurface = new SDL::Surface( -name => "ball.png" );
  $DarkGreenBallSurface =  new SDL::Surface( -name => "ball.png" );
  $GrayBallSurface = new SDL::Surface( -name => "ball.png" );
  $DarkGrayBallSurface =  new SDL::Surface( -name => "ball.png" );
  $GoldBallSurface = new SDL::Surface( -name => "ball.png" );
  $DeathBallSurface = new SDL::Surface( -name => "ball.png" );
  
  $HexaSurface = new SDL::Surface( -name => "hexa.png" );
  $DarkHexaSurface = new SDL::Surface( -name => "hexa.png" );
  $BorderSurface = new SDL::Surface( -name => "border.png" );
  $BonusSurface = new SDL::Surface( -name => "bonus.png" );
  
  my $srcrect = new SDL::Rect(-w=>32, -h=>32);
  $BonusSurface->blit($srcrect, $GreenBallSurface, new SDL::Rect(-x=>48, -y=>48));
  $BonusSurface->blit($srcrect, $GreenBallSurface, new SDL::Rect(-x=>160, -y=>32));
  $BonusSurface->blit($srcrect, $GreenBallSurface, new SDL::Rect(-x=>16, -y=>144));
  $BonusSurface->blit($srcrect, $DarkGreenBallSurface, new SDL::Rect(-x=>48, -y=>48));
  $BonusSurface->blit($srcrect, $DarkGreenBallSurface, new SDL::Rect(-x=>160, -y=>32));
  $BonusSurface->blit($srcrect, $DarkGreenBallSurface, new SDL::Rect(-x=>16, -y=>144));
  $srcrect->x(32);
  $BonusSurface->blit($srcrect, $GoldBallSurface, new SDL::Rect(-x=>48, -y=>48));
  $BonusSurface->blit($srcrect, $GoldBallSurface, new SDL::Rect(-x=>160, -y=>32));
  $BonusSurface->blit($srcrect, $GoldBallSurface, new SDL::Rect(-x=>16, -y=>144));
  
  &AlterPalette( $DarkBallSurface, sub { 1; }, sub { return $_[1] * 2 / 3, $_[2] * 2 / 3, $_[3] * 2 / 3; } );
  &AlterPalette( $DarkHexaSurface, sub { 1; }, sub { return $_[1] * 2 / 3, $_[2] * 2 / 3, $_[3] * 2 / 3; } );
  &AlterPalette( $GreenBallSurface, sub { 1; }, 
    sub { shift @_; my ($h, $s, $i) = &RgbToHsi(@_); return &HsiToRgb( 85, $s, $i ); } );
  &AlterPalette( $DarkGreenBallSurface, sub { 1; }, 
    sub { shift @_; my ($h, $s, $i) = &RgbToHsi(@_); my ($r,$g,$b) = &HsiToRgb( 85, $s, $i); return ($r*2/3, $g*2/3, $b*2/3); } );
  &AlterPalette( $GoldBallSurface, sub { 1; }, 
    sub { shift @_; my ($h, $s, $i) = &RgbToHsi(@_); return &HsiToRgb( 35, 255, $i ); } );
  &AlterPalette( $DeathBallSurface, sub { 1; }, 
    sub { my ($i) = int($_[1] + $_[2] + $_[3]) / 6; return ($i, $i, $i) } );
  
  &MakeGuySurfaces();
}

sub MakeGuySurfaces {
  foreach my $player (@Players) {
    my $guySurface = new SDL::Surface( -name => "guy.png" );
    my $harpoonSurface = new SDL::Surface( -name => "harpoon.png" );
    
    &AlterPalette( $guySurface, sub { $_[3] > $_[2] and $_[3] > $_[1]; },
      sub {
        shift @_;
        my ($h, $s, $i) = &RgbToHsi(@_);
        return &HsiToRgb($player->{hue}, $player->{saturation}, $i); } 
    );
    &AlterPalette( $harpoonSurface, sub { 1; },
      sub {
        shift @_;
        my ($h, $s, $i) = &RgbToHsi(@_);
        return &HsiToRgb($player->{hue}, $s, $i); } 
    );
    $player->{guySurface} = $guySurface;
    $player->{harpoonSurface} = $harpoonSurface;
  }
}

sub AlterPalette {
  my ($surface, $filterSub, $alterSub) = @_;
  my ($r, $g, $b);
  my ($palette, $numColors, $n, $color);
  
  $palette = $surface->palette();
  $numColors = SDL::PaletteNColors($palette);
  for ($n = 0; $n < $numColors; ++$n) {
    $color = SDL::PaletteColors($palette, $n);
    ($r, $g, $b) = ( SDL::ColorR($color), SDL::ColorG($color), SDL::ColorB($color) );
    
    next unless $filterSub->($n, $r, $g, $b);
    ($r, $g, $b) = $alterSub->($n, $r, $g, $b);
    
    SDL::PaletteColors($palette, $n, $r, $g, $b);
  }
}

sub RenderBorder {
  my ($borderSurface, $targetSurface, $x, $y, $width, $height) = @_;
  my ($dstrect, $srcrect1, $srcrect2, $xpos, $ypos);
  
  # Draw top and bottom border
  
  $dstrect = new SDL::Rect( -width => 32, -height =>16);
  $srcrect1 = new SDL::Rect( -width => 32, -height =>16, -x => 32, -y => 0);
  $srcrect2 = new SDL::Rect( -width => 32, -height =>16, -x => 32, -y => 80);
  for ($xpos = 32; $xpos < $width-32; $xpos += 32) {
    $dstrect->x($xpos + $x);
    $dstrect->y($y);
    $borderSurface->blit($srcrect1, $targetSurface, $dstrect);
    $dstrect->y($y + $height - 16);
    $borderSurface->blit($srcrect2, $targetSurface, $dstrect);
  }
  
  # Draw left and right border
  $dstrect->width(16);
  $dstrect->height(32);
  $srcrect1 = new SDL::Rect( -width => 16, -height =>32, -x => 0, -y => 32);
  $srcrect2 = new SDL::Rect( -width => 16, -height =>32, -x => 80, -y => 32);
  for ($ypos = 32; $ypos < $height-32; $ypos += 32) {
    $dstrect->x($x);
    $dstrect->y($ypos + $y);
    $borderSurface->blit($srcrect1, $targetSurface, $dstrect);
    $dstrect->x($x + $width - 16);
    $borderSurface->blit($srcrect2, $targetSurface, $dstrect);
  }
  
  # Draw the borders
  $srcrect1 = new SDL::Rect( -width => 32, -height => 32 );
  $dstrect->width(32);
  $dstrect->x($x);
  $dstrect->y($y);
  $borderSurface->blit($srcrect1, $targetSurface, $dstrect);
  $srcrect1->x(64);
  $dstrect->x($x + $width - 32);
  $borderSurface->blit($srcrect1, $targetSurface, $dstrect);
  $srcrect1->y(64);
  $dstrect->y($y + $height - 32);
  $borderSurface->blit($srcrect1, $targetSurface, $dstrect);
  $srcrect1->x(0);
  $dstrect->x($x);
  $borderSurface->blit($srcrect1, $targetSurface, $dstrect);
}

sub LoadBackground {
  my $filename = shift;
  my ($backgroundImage, $srcrect, $dstrect);
  
  $Background->fill(  new SDL::Rect(-width=>$PhysicalScreenWidth, -height=>$PhysicalScreenHeight), new SDL::Color() );
  $backgroundImage = new SDL::Surface(-name => $filename);
  $dstrect = new SDL::Rect(-x => $ScreenMargin, -y => $ScreenMargin);
  $srcrect = new SDL::Rect(-width => $ScreenWidth, -height => $ScreenHeight);
  $backgroundImage->blit($srcrect, $Background, $dstrect);
  
  &RenderBorder($BorderSurface, $Background, 0, 0, $PhysicalScreenWidth, $ScreenHeight + 2 * $ScreenMargin);
}

##########################################################################
# SOUNDS
##########################################################################

sub LoadSounds {
  $Mixer = eval { SDL::Mixer->new(-frequency => 22050, -channels => 2, -size => 1024); };
  if ($@) {
    warn $@;
	return 0;
  }
  
  my ($soundName, $fileName);
  while (($soundName, $fileName) = each %Sounds) {
    $Sounds{$soundName} = new SDL::Sound($fileName);
  }
  
  $::music = new SDL::Music('UPiPang.mp3');
  $::music = new SDL::Music('UPiPang.mid') unless $::music->{-data};
  $Mixer->play_music($::music, -1);
}

sub PlaySound {
  my $sound = shift;
  $Mixer and $Sounds{$sound} and $Mixer->play_channel(-1, $Sounds{$sound}, 0);
}


##########################################################################
# GAME MANAGEMENT
##########################################################################

sub ResetGame {
  &SetGameLevel(0);
  $PlayersAlive = $NumGuys;
  $GamePause = 0;
  $SpawnDelay = 0;
  $SuperBallDelay = 2500 + rand(2500);  # 25sec - 50sec
  @GameObjects = ();
  my @guys = ();
  for (my $i=0; $i < $NumGuys; ++$i) {
    my $player = $Players[$i];
    $player->{score} = 0;
    $player->{scoreforbonuslife} = 200000;
    $player->{lives} = 2;
    $player->{startX} = ($ScreenWidth - $NumGuys * 50) / 2 + 60 * ($i+0.5) - 32;
    my $guy = new Guy($i);
    push @guys, ($guy);
  }
  push @GameObjects, (@guys);
}

sub SetGameLevel {
  my $level = shift;
  
  $GameLevel = $level;
  $GameLevelDesc = $GameLevels[$GameLevel >= scalar @GameLevels ? $#GameLevels : $GameLevel];
  $GameSpeed = $GameLevelDesc->{gamespeed} * 0.8;
  $GameLevelAdvance = 0;
}

sub CalculateAdvances {
  my $advance = &GameTimer::GetAdvances();
  if ($advance <= 0) {
    $App->delay(10);
    $advance = &GameTimer::GetAdvances();
  } elsif ($advance > 10) {
    print STDERR "advance = $advance!\n";
    $advance = 10;
  }
  return $advance;
}

sub AdvanceGame {
  foreach my $gameObject (@GameObjects) {
    $gameObject->Advance();
  }
  &SpawnBalls() if $GamePause <= 0;
  --$GamePause if $GamePause > 0;
}

sub SpawnBalls {
  my ($randmax, $rnd, $ballName, $balldesc);
  
  --$SuperBallDelay;
  if ($SuperBallDelay <= 0) {
    push @GameObjects, (
      new DarkBall($BallDesc[rand(2) + 1], rand(600), rand(40) < 20 ? 0 : 1, 1) );
    $SuperBallDelay = 5000 + rand(5000); # 50sec - 100sec
  }

  --$SpawnDelay;
  return if $SpawnDelay > 0;
  $randmax = 1000;
  while ($SpawnDelay <= 0) {
    $rnd = int(rand($randmax));
    $randmax = 0;
    
    # We try to find the balldesc that falls at $rnd
    my $ballRoulette = $GameLevelDesc->{balls};
    for (my $i = 0; $i < scalar @{$ballRoulette}; $i+=2) {
      my $rouletteWeight = $ballRoulette->[$i+1];
      $randmax += $rouletteWeight;
      $rnd -= $rouletteWeight;
      if ($rnd < 0) {
        $ballName = $ballRoulette->[$i];
        last;
      }
    }
    next unless ($ballName); # $rnd too large.. We'll have a better $randmax this time!
    
    ($balldesc) = grep( $_->{name} eq $ballName, @BallDesc);
    last if $balldesc;
  }
  
  push @GameObjects, ( new DarkBall($balldesc, rand(600), rand(40) < 20 ? 0 : 1) );
  $SpawnDelay = $GameLevelDesc->{spawndelay} * $balldesc->{spawndelay} * 70;
  $SpawnDelay /= ($NumGuys + 1) / 2;
}

sub OnBallPopped {
  ++$GameLevelAdvance;
  if ($GameLevelAdvance >= 18) {
    &PlaySound('level');
    &SetGameLevel($GameLevel+1);
  }
}

sub SuperKill {
  my $guy = shift;
  
  my @gameObjects = @GameObjects;
  my $sound = 0;
  foreach my $ball (@gameObjects) {
    next unless $ball->isa("Ball");
    $ball->Pop($guy, 'superkill');
    $sound = 1;
  }
  &PlaySound('pop') if $sound;
}

sub PlayerKilled {
  my (@killedGuys, @deadGuys, $guy, $i);
  
  &DrawGame();
  &PlaySound('death');
  $::App->delay(1000);
  &GameTimer::ResetTimer();
  
  @killedGuys = grep { $_->{justkilled}; } @GameObjects;
  foreach $guy (@killedGuys) {
    $guy->Clear();
    $guy->{killed} = 1;
    push @deadGuys, (new DeadGuy($guy));
  }
  push @GameObjects, (@deadGuys);
  
  for ($i = 0; $i < 300; ++$i) {
   &HandleEvents();
   my $advance = &CalculateAdvances();
    while ($advance--) {
      foreach my $gameObject (@deadGuys) {
        $gameObject->Advance();
      }
    }
    &DrawGame();
    last if $deadGuys[0]->{deleted};
  }
  
  foreach $guy (@killedGuys) {
    $guy->DeleteHarpoons;
    if ($guy->{desc}->{lives}--) {
      $guy->{x} = $guy->{desc}->{startX};
      $guy->{invincible} = 500;
      $guy->{killed} = 0;
      $guy->{justkilled} = 0;
    } else {
      # One player less
      $guy->Delete();
      --$PlayersAlive;
    }
  }
  &GameTimer::ResetTimer();
}

sub HandleEvents {
  my ($event, $type);

  $event = new SDL::Event;
  while (1) {
    last unless $event->poll();
    $type = $event->type();
   
    if ($type == SDL_QUIT) {
      exit;
    }
    elsif ($type == SDL_KEYDOWN) {
      my $keypressed = $event->key_sym;
      $Keys{$keypressed} = 1;
      $Events{$keypressed} = 1;
      if ($keypressed == SDLK_ESCAPE) {
        exit;
      }
      if ($keypressed == SDLK_w) {
        $::App->save_bmp("screenshot.bmp");
      }
      if ($keypressed == SDLK_p) {
        $::Keys = (); $::Events = ();
        while (1) {
          $event->wait();
          last if $event->type() == SDL_KEYDOWN and $event->key_sym == SDLK_p;
          exit if $event->type() == SDL_KEYDOWN and $event->key_sym == SDLK_ESCAPE;
        }
        &GameTimer::ResetTimer();
      }
    }
    elsif ($type == SDL_KEYUP) {
      my $keypressed = $event->key_sym;
      $Keys{$keypressed} = 0;
    }
  }
  
  &Joystick::ReadJoystick();
}

##########################################################################
# GAME DRAWING
##########################################################################

sub DrawGame {
  my (%options) = @_;
  
  my ($gameObject);
  foreach $gameObject (@GameObjects) {
    $gameObject->Clear();
  }  
  foreach $gameObject (@GameObjects) {
    $gameObject->Draw();
  }
  
  unless ($options{-noscore}) {
    my ($x, $y, $widthPerGuy);
    $x = 10;
    $y = $PhysicalScreenHeight - 70;
    $App->fill( new SDL::Rect(-x=>0, -y=>$y, -width=>$PhysicalScreenWidth, -height=>$PhysicalScreenHeight - $y), new SDL::Color() );
    &DrawLevelIndicator( $x, $y );
    
    $widthPerGuy = ($PhysicalScreenWidth - 110) / $NumGuys ;
    for (my $i=0; $i < $NumGuys; ++$i) {
      $x = $PhysicalScreenWidth - 120 - $widthPerGuy * ($NumGuys - $i - 1);
      &DrawScore( $Players[$i], $x, $y );
    }
  }

  $App->sync();
}

sub DrawLevelIndicator {
  my ($x, $y) = @_;
  $App->print( $x, $y, 'Level ' . ($GameLevel+1) );
  $App->print( $x, $y + 24, int(100 * $GameLevelAdvance / 18) . '%' );
  $App->print( $x, $y + 48, 'spd: ' . int($GameLevelDesc->{gamespeed} * 100) . '/' . int($GameLevelDesc->{spawndelay}) );
}

sub DrawScore {
  my ($player, $x, $y) = @_;
  my ($i, $srcrect, $dstrect, $scoreText);
  
  $scoreText = sprintf($player->{score});
  $srcrect = new SDL::Rect(-width => 16, -height => 16, -y => 224);
  $dstrect = new SDL::Rect(-width => 16, -height => 16, -y => $y, -x => $x);
  for ($i = 0; $i < length($scoreText); ++$i) {
    $srcrect->x(320 + (ord(substr($scoreText, $i)) - ord('0')) * 16);
    $dstrect->x($x + $i * 16);
    $player->{guySurface}->blit( $srcrect, $App, $dstrect );
  }
  $srcrect = new SDL::Rect(-width => 32, -height => 32, -x =>320, -y => 192);
  $dstrect = new SDL::Rect(-width => 32, -height => 32, -x =>$x, -y => $y + 24);
  foreach $i ( 0 .. $player->{lives}-1 ) {
    $dstrect->x( $x + $i * 36 );
    $player->{guySurface}->blit( $srcrect, $App, $dstrect );
  }
}


sub MenuAdvance {
  my $advance = &CalculateAdvances();
  %Events = ();
  %GameEvents = ();
  &HandleEvents();
  while ($advance--) { 
    $SpawnDelay = 10;
    $SuperBallDelay = 10;
    &AdvanceGame(); 
  }
  while (ref($GameObjects[$#GameObjects]) ne 'MenuItem') { unshift @GameObjects, (pop @GameObjects); }
  &DrawGame(-noscore => 1);
}

sub UpdateMenu {
  my (@menuItems) = @_;
  
  $menuItems[2]->SetText("< Number of Players: $NumGuys >");
  for ( my $i=3; $i <= 8; ++$i) {
    $menuItems[$i]->{x} = $i-3 < $NumGuys ? 50 : $PhysicalScreenWidth * 2;
  }
}

sub Menu {
  my ($oldBallSurface, $oldHexaSurface) = ($BallSurface, $HexaSurface);
  my ($menuFont, @menuItems, $currentItemIndex, $currentItem);
  
  ($BallSurface, $HexaSurface) = ($DarkBallSurface, $DarkHexaSurface);
  $Background->blit(0, $App, 0);
  $App->print( 10, $ScreenHeight + 40, "Use cursor keys to navigate menu, Enter to select" );
  $App->print( 10, $ScreenHeight + 60, "P pauses the game, Esc quits" );
  $App->print( 10, $ScreenHeight + 80, "Pang Zero $::Version Copyright (C) 2006 by UPi (upi\@sourceforge.net)" );
  $menuFont = new SDL::Font("font2.png");
  @GameObjects = ();
  $GameSpeed = 1.0;
  $GamePause = 0;
  
  @menuItems = (
    new MenuItem( 300,  70, "PANG ZERO" ),
    new MenuItem( 100, 140, "Start Game" ),
    new MenuItem(  68, 180, "<>"),
  );
  
  for ( my $i = 1; $i <= 6; ++$i ) {
    push @menuItems, (new MenuItem( 50, 210 + $i * 35, "Setup Player $i"));
  }
  &UpdateMenu(@menuItems);
  
  push @GameObjects, (
    new DarkBall($BallDesc[8], rand(600), 1), 
    new DarkBall($BallDesc[0], rand(600), 0), 
    new DarkBall($BallDesc[1], rand(600), 1, 1),
    new DarkBall($BallDesc[2], rand(600), 0),
    new DarkBall($BallDesc[5], rand(600), 1),
    @menuItems,
  );
  
  $menuItems[0]->{filled} = 1;
  $menuItems[0]->{fillcolor} = new SDL::Color(-b=>255, -g=>128);
  $currentItemIndex = 1;
  $currentItem = $menuItems[$currentItemIndex];
  $currentItem->Select;
  
  &GameTimer::ResetTimer();
  
  while (1) {
    &MenuAdvance();
    if ($Events{SDLK_DOWN()}) {
      ++$currentItemIndex unless $currentItemIndex >= $NumGuys + 2;
      $currentItem = $menuItems[$currentItemIndex];
      $currentItem->Select;
    }
    if ($Events{SDLK_UP()}) {
      --$currentItemIndex unless $currentItemIndex <= 1;
      $currentItem = $menuItems[$currentItemIndex];
      $currentItem->Select;
    }
    if ($Events{SDLK_LEFT()} and $currentItemIndex == 2) {
      --$NumGuys if $NumGuys > 1;
      &UpdateMenu(@menuItems);
#     $currentItem->SetText("< Number of Players: $NumGuys >");
    }
    if ($Events{SDLK_RIGHT()} and $currentItemIndex == 2) {
      ++$NumGuys if $NumGuys < 6;
      &UpdateMenu(@menuItems);
#      $currentItem->SetText("< Number of Players: $NumGuys >");
    }
    if ($Events{SDLK_RETURN()}) {
      last if $currentItemIndex == 1; # Start game
      next if $currentItemIndex == 2; # Ignored
      my $player = $Players[$currentItemIndex - 3];
      my $key = 0;
      my @prompts = ("Press 'LEFT' key", "Press 'RIGHT' key", "Press 'FIRE' key");
      my $keyMenuItem = new MenuItem( 300, $currentItem->{y}, $prompts[0] );
      push @GameObjects, ($keyMenuItem);
      $keyMenuItem->Select;
      while (1) {
        &MenuAdvance();
        if (%Events) {
          ($player->{keys}->[$key]) = (%Events);
          ++$key;
          last if $key >= 3;
          $keyMenuItem->SetText($prompts[$key]);
        }
      }
      $currentItem->Select;
      $keyMenuItem->Delete;
    }
  }
  
  ($BallSurface, $HexaSurface) = ($oldBallSurface, $oldHexaSurface);
  $ScoreFont->use();
}


##########################################################################
# MAIN PROGRAM STARTS HERE
##########################################################################

my $sdlFlags = SDL_ANYFORMAT; # | SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_HWACCEL | SDL_ASYNCBLIT;

$App = new SDL::App
  -flags => $sdlFlags,
  -title => "Pang Zero $::Version", 
  -width => $PhysicalScreenWidth, 
  -height => $PhysicalScreenHeight,
  -fullscreen => 1,
;

&SDL::ShowCursor(0);

$Background = new SDL::Surface( 
  -name =>'',
  -flags=>SDL_SWSURFACE,
  -width => $App->width,
  -height => $App->height,
  -depth => 32,
  -Amask => '0 but true');

$ScoreFont = new SDL::Font("brandybun3.png");

&LoadSurfaces();
&LoadSounds();
&Joystick::InitJoystick();

my $screenRect = new SDL::Rect(-width=>$PhysicalScreenWidth, -height=>$PhysicalScreenHeight);
my $blankColor = new SDL::Color();



StartGameOver:

$ScreenHeight = int( ($PhysicalScreenHeight - $ScreenMargin * 2 - 60) / 32 ) * 32;
&RenderBorder($BorderSurface, $Background, 0, 0, $PhysicalScreenWidth, $ScreenHeight + 2 * $ScreenMargin);
&LoadBackground('desert2.png');
$Background->blit(0, $App, 0);

&Menu();

$ScreenHeight = int( ($PhysicalScreenHeight - $ScreenMargin * 2 - 150) / 32 ) * 32;
&RenderBorder($BorderSurface, $Background, 0, 0, $PhysicalScreenWidth, $ScreenHeight + 2 * $ScreenMargin);
&LoadBackground('desert2.png');
$Background->blit(0, $App, 0);
&ResetGame();
&GameTimer::ResetTimer();

push @GameObjects, (
#  new DarkBall($BallDesc[11], 300, 0),
#  new DarkBall($BallDesc[8], 400, 1),
);
#$SpawnDelay = 1000;

my $superKillCount = 0;
my $superKillDelay = 0;
my $superKillGuy;

while (1) {
  
  # Calculate advance (how many game updates to perform)
  my $advance = &CalculateAdvances();
  
  # Advance the game
  
  %Events = ();
  &HandleEvents();
  while ($advance--) {
  
    if ($superKillCount > 0) {
      if (--$superKillDelay <= 0) {
         --$superKillCount;
         $superKillDelay = 50;
         &SuperKill($superKillGuy);
      }
    }
  
    %GameEvents = ();
    &AdvanceGame();
    if ($GameEvents{kill} ) {
      &PlayerKilled();
      last if $PlayersAlive <= 0;
      $GamePause = 200 if $GamePause < 200;
      &GamePause::Show();
    } elsif ($GameEvents{magic}) {
      if ($GamePause < 200) { 
        $GamePause = 200; &PlaySound('pause'); 
        &GamePause::Show();
      }
    } elsif ($GameEvents{superpause}) {
      if ($GamePause < 800) {
        $GamePause = 800; &PlaySound('pause'); 
        &GamePause::Show();
      }
    } elsif ($GameEvents{superkill}) {
      $superKillCount = 5;
      $superKillDelay = 0;
      $superKillGuy = $GameEvents{superkill};
      $SpawnDelay = 250;
      $SuperBallDelay += 1000;
      my @gameObjects = @GameObjects;
      foreach my $darkBall (@gameObjects) { $darkBall->Delete if ref($darkBall) eq 'DarkBall'; }
    }
  }
  last if $PlayersAlive <= 0;
  &DrawGame();
}

$App->delay(2000);
goto StartGameOver;
